# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from oslo_config import cfg
from oslo_config.sources import _environment


class EnvSource(_environment.EnvironmentConfigurationSource):
    """
    EnvironmentConfigurationSource with custom prefix
    """
    prefix = ""

    @classmethod
    def _make_name(cls, group_name, option_name):
        group_name = group_name or 'PYJAMAAS'
        return '{}{}_{}'.format(
            cls.prefix, group_name.upper(), option_name.upper())


config = cfg.ConfigOpts()
config._env_driver = EnvSource()

grp = cfg.OptGroup('maas')
opts = [
    cfg.StrOpt(
        'api_url',
        required=True,
        help='MaaS API URL',
        sample_default='https://maas:5240/MAAS/api/2.0'
    ),
    cfg.StrOpt(
        'api_key',
        required=True,
        help='MaaS API Key',
        sample_default='AAAAAAAAAA:BBBBBBBBBBBBB:CCCCCCCCCCCCC'
    ),
]

config.register_group(grp)
config.register_opts(opts, group=grp)


def list_opts():
    """
    used as entrypoing for `oslo-config-generator`
    """
    return [
        [g, [o['opt'] for o in g._opts.values()]]
        for g in config._groups.values()]



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys

from cliff.command import Command

from pyjamaas import parse, maas
from pyjamaas.log import log


class Clone(Command):
    """
    clone configuration
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--source', type=parse.yaml_argument, required=True,
            help='source machine')
        parser.add_argument(
            '--destination', type=parse.yaml_argument, required=True,
            help='destination machines')
        parser.add_argument(
            '--force', action='store_true', help='do not ask for confirmation')
        parser.add_argument(
            '--network', action='store_true', help='clone interface config')
        parser.add_argument(
            '--storage', action='store_true', help='clone storage config')
        return parser

    def take_action(self, parsed_args):
        source = maas.match_machines(parsed_args.source)
        if len(source) > 1:
            log.fatal('Multiple source machines were matched')
            sys.exit(1)

        system_id, hostname = source[0]['system_id'], source[0]['hostname']
        log.info(f'Using {system_id} ({hostname}) as source')

        destination = maas.match_machines(parsed_args.destination)

        # filter out source machine from destinations
        destination = [d for d in destination if d['system_id'] != system_id]

        log.info(f'Will apply configuration to {len(destination)} machines:')
        for machine in destination:
            m_id, m_name = machine['system_id'], machine['hostname']
            log.info(f'- {m_id} ({m_name})')

        if not parsed_args.force:
            reply = input('Are you sure [y/N]? ')
            if reply.lower() != 'y':
                log.error('Aborted due to user input')

        s = maas.session()
        log.info('Cloning')
        r = s.Machines.clone(
            system_id=system_id,
            destinations=destination,
            interfaces=parsed_args.network,
            storage=parsed_args.storage
        )
        log.info(f'Result: {r}')



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from itertools import chain
import sys

from cliff.app import App
from cliff.commandmanager import CommandManager

from pyjamaas.config import config

README = 'https://github.com/grnet/pyjamaas/blob/master/README.md'


class PyjamaasApp(App):
    def __init__(self):
        super(PyjamaasApp, self).__init__(
            description='pyjamaas: Python MaaS and Juju Toolkit',
            version='0.3',
            command_manager=CommandManager('pyjamaas.entrypoints'),
            deferred_help=True
        )

    def build_option_parser(self, description, version, argparse_kwargs=None):
        parser = super().build_option_parser(
            description, version, argparse_kwargs=argparse_kwargs)

        parser.add_argument(
            '--config-file', nargs='+', help='configuration file', metavar='PATH')
        parser.add_argument(
            '--config-dir', nargs='+', help='configuration directory', metavar='DIR')

        parser.epilog = f'Refer to {README} for configuration instructions'
        return parser

    def initialize_app(self, argv):
        super().initialize_app(argv)

        args = []
        if self.options.config_file is not None:
            args.extend(chain.from_iterable(
                ['--config-file', cfg] for cfg in self.options.config_file))
        if self.options.config_dir is not None:
            args.extend(chain.from_iterable(
                ['--config-dir', cfg] for cfg in self.options.config_dir))

        config(project='pyjamaas', args=args)


def main(argv=sys.argv[1:]):
    m = PyjamaasApp()
    return m.run(argv)


if __name__ == '__main__':
    sys.exit(main())



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from cliff.lister import Lister
from cliff.show import ShowOne

from pyjamaas import parse, maas, util


class List(Lister):
    """
    list MaaS machines
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parse.machines_arguments(parser)
        parser.add_argument('--all', action='store_true', help='show all info')
        return parser

    def take_action(self, parsed_args):
        return util.list_dict_to_cliff(maas.match_machines(
            parsed_args.match, formatted=not parsed_args.all))


class Get(ShowOne):
    """
    get info for a MaaS machine
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parse.machines_arguments(parser)
        parser.add_argument('--all', action='store_true', help='show all info')
        return parser

    def take_action(self, parsed_args):
        return util.dict_to_cliff(maas.match_machines(
            parsed_args.match, formatted=not parsed_args.all)[0])



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import yaml


def yaml_argument(arg):
    """
    argparse YAML argument type
    """
    try:
        if arg is not None and arg.startswith('@'):
            with open(arg[1:], 'r') as fin:
                return yaml.safe_load(fin)
        return yaml.safe_load(arg)
    except yaml.YAMLError as e:
        raise argparse.ArgumentTypeError('invalid YAML') from e
    except OSError as e:
        raise argparse.ArgumentTypeError('could not load file') from e


def maas_arguments(parser):
    """
    MaaS configuration
    """


def machines_arguments(parser):
    """
    Arguments for matching against MaaS machines
    """
    parser.add_argument(
        '--match', type=yaml_argument, required=True,
        help='Choose MaaS machines')



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


def list_dict_to_cliff(data):
    """
    convert list of dicts to Cliff (columns, (values)) format
    """
    if not data:
        return (None, None)

    columns = list(data[0].keys())
    values = list([d[col] for col in columns] for d in data)

    return columns, values


def dict_to_cliff(data):
    """
    convert list of dicts to Cliff (columns, (values)) format
    """
    columns, values = list_dict_to_cliff([data])
    return columns, values[0]



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from maas.client.bones import SessionAPI, CallError, helpers

from pyjamaas.config import config
from pyjamaas.log import log


__session = None


# Collection of errors that may happen when using MaaS API
APIError = (CallError, helpers.RemoteError)


def session():
    """
    Return a sessions to the MaaS server. Subsequent calls reuse the same
    connection
    """
    global __session
    if __session:
        return __session

    if not config.maas.api_url:
        raise RuntimeError('MAAS_API_URL is not set, see `pyjamaas --help`')
    if not config.maas.api_key:
        raise RuntimeError('MAAS_API_KEY is not set, see `pyjamaas --help`')

    log.info(f'Connecting to {config.maas.api_url}')
    _, __session = SessionAPI.connect(
        config.maas.api_url, apikey=config.maas.api_key)

    return __session


def format_machine(machine):
    """
    format/filter machine info from MaaS
    """
    return {
        'hostname': machine['hostname'],
        'system_id': machine['system_id'],
        'status': machine['status_name'],
        'tags': machine['tag_names'],
    }


def _get(options, key):
    s = options.get(key) or []
    if not isinstance(s, list):
        s = [s]
    return s


def match_machines(options, formatted=True):
    """
    Match machines based on tags, hostname, or system id
    """
    s = session()

    # match with hostname by default
    if isinstance(options, str):
        options = {'hostname': options}

    log.debug(f'Looking for machines with filter {options}')

    match_tags = _get(options, 'tags')
    match_hostnames = _get(options, 'hostname')
    match_system_ids = _get(options, 'system_id')

    if match_system_ids:
        machines = s.Machines.read(id=match_system_ids)
    elif match_hostnames:
        machines = s.Machines.read(hostname=match_hostnames)
    elif match_tags:
        unfiltered_machines = s.Tag.machines(name=match_tags[0])
        machines = [m for m in unfiltered_machines if all(
            t in m['tag_names'] for t in match_tags)]
    else:
        log.error('Not fetching machines with empty filter')
        machines = []

    if formatted:
        machines = [format_machine(m) for m in machines]

    return machines



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import defaultdict
import sys

from cliff.command import Command

from pyjamaas import parse, maas
from pyjamaas.log import log


class Add(Command):
    """
    add tags
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parse.machines_arguments(parser)
        parser.add_argument(
            '--tags', required=True, nargs='+', metavar='TAG',
            help='list of tags')
        parser.add_argument(
            '--no-create', action='store_true',
            help='do not create tags automatically if they do not exist')
        return parser

    def take_action(self, parsed_args):
        s = maas.session()
        existing_tags = [t['name'] for t in s.Tags.read()]

        new_tags = [t for t in parsed_args.tags if t not in existing_tags]
        if new_tags and parsed_args.no_create:
            log.fatal('Will not create new tags with --no-create flag')
            sys.exit(1)

        for tag in new_tags:
            log.info(f'Creating new tag {tag}')
            r = s.Tags.create(name=tag, description=f'{tag} (pyjamaas)')
            log.debug(f'Result: {r}')

        machines = maas.match_machines(parsed_args.match)
        system_ids = [m['system_id'] for m in machines]

        for tag in parsed_args.tags:
            log.info(f'Adding tag {tag} to {len(machines)} machines')
            r = s.Tag.update_nodes(name=tag, add=system_ids)
            log.debug(f'Result: {r}')


class Clear(Command):
    """
    clear tags from machines
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parse.machines_arguments(parser)
        g = parser.add_mutually_exclusive_group(required=True)
        g.add_argument('--all-tags', action='store_true', help='clear all')
        g.add_argument('--tags', nargs='+', metavar='TAG', help='list of tags')
        return parser

    def take_action(self, parsed_args):
        machines = maas.match_machines(parsed_args.match)
        s = maas.session()

        clear_tags = defaultdict(lambda: [])
        for machine in machines:
            if parsed_args.all_tags:
                parsed_args.tags = machine['tags']

            for tag in set(parsed_args.tags) & set(machine['tags']):
                clear_tags[tag].append(machine['system_id'])

        for tag, system_ids in clear_tags.items():
            log.info(f'Clearing tag {tag} from {len(system_ids)} machines')
            r = s.Tag.update_nodes(name=tag, remove=system_ids)
            log.debug(r)


class Delete(Command):
    """
    delete a tag
    """
    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'tags', nargs='+', metavar='TAG', help='list of tags to delete')
        return parser

    def take_action(self, parsed_args):
        s = maas.session()
        for tag in parsed_args.tags:
            log.info(f'Deleting tag {tag}')
            r = s.Tag.delete(name=tag)
            log.debug(f'Response: {r}')



# Copyright (C) 2020  GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re
from setuptools import setup, find_packages


def get_file(name):
    try:
        with open(name, 'r') as fin:
            return fin.read()
    except OSError:
        return ""


def get_version():
    changelog = get_file('CHANGELOG.md')
    version = next(re.finditer(r'## \[v([\d.]*)\]', changelog), None)
    return version and version.group(1) or 'none'


def setup_package():
    setup(
        name='pyjamaas',
        version=get_version(),
        description='Python MaaS and Juju toolkit',
        long_description=get_file('README.md'),
        long_description_content_type='text/markdown',
        url='https://github.com/grnet/pyjamaas.git',
        packages=find_packages(),
        entry_points={
            'console_scripts': [
                'pyjamaas = pyjamaas.cliff:main',
            ],
            'pyjamaas.entrypoints': [
                'clone = pyjamaas.commands.clone:Clone',
                'config = pyjamaas.commands.config:Config',
                'machine list = pyjamaas.commands.machine:List',
                'machine get = pyjamaas.commands.machine:Get',
                'tags add = pyjamaas.commands.tags:Add',
                'tags clear = pyjamaas.commands.tags:Clear',
                'tags delete = pyjamaas.commands.tags:Delete',
            ],
            'oslo.config.opts': [
                'pyjamaas = pyjamaas.config:list_opts',
            ]
        },
        install_requires=[
            'cliff',
            'oslo.config',
            'python-libmaas',
            'pyyaml',
        ],
        classifiers=[
            'Intended Audience :: System Administrators',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Operating System :: OS Independent',
        ],
    )


if __name__ == '__main__':
    setup_package()
